@@ -465,7 +465,12 @@ def sql_scheduler(
     w_print(f"Execution Complete.")
     print(
         construct_table(
-            ["task_id", "script duration (s)", "test duration (s)"],
+            [
+                "task_id",
+                "script duration (s)",
+                "test duration (s)",
+                "upstream test duration (s)",
+            ],
             [
                 (
                     task.task_id.lower(),
@@ -485,6 +490,14 @@ def sql_scheduler(
                             1,
                         )
                     ),
+                    str(
+                        round(
+                            task.upstream_test_duration
+                            if task.upstream_test_duration is not None
+                            else -1,
+                            1,
+                        )
+                    ),
                 )
                 for task in sorted(
                     filter(
